// cpu.h
// Revision 18-jan-2004

// Testing with two z80 emualtions: libz80 and ImcZ80, taken from
// the Joyce Amstrad PCW emulator, which in turn is based in xz80.

// Working with ImcZ80 but disassembly with libz80
#define DISASM_WITH_libz80

#if ! defined USE_ImcZ80 || defined DISASM_WITH_libz80

extern "C" {
#include "z80.h"
}

#endif

#ifdef USE_ImcZ80

#include "Z80.hxx"

#endif

#include <vector>
#include <ostream>

#ifndef USE_ImcZ80
class Cpu {
#else
class Cpu : private ImcZ80 {
#endif

private:
	int self;
	static std::vector <Cpu *> vi;
	static int putvi (Cpu *);
protected:
	byte mem [65536];
private:
	static byte memread (int param, ushort address);
	static void memwrite (int param, ushort address, byte data);
	static byte ioread (int param, ushort address);
	static void iowrite (int param, ushort address, byte data);
public:
	byte memread (ushort address);
	ushort memreadw (ushort address);
	void memwrite (ushort address, byte data);
	void memwritew (ushort address, ushort data);

	#if ! defined USE_ImcZ80 || defined DISASM_WITH_libz80

private:
	Z80Context z;

	#endif

	#ifndef USE_ImcZ80

public:
	static byte flag_c (byte f) { return f & F_C; }
	static byte flag_z (byte f) { return f & F_Z; }
	static byte flag_s (byte f) { return f & F_S; }

	byte get_a () { return z.R1.br.A; }
	byte get_f () { return z.R1.br.F; }
	byte get_b () { return z.R1.br.B; }
	byte get_c () { return z.R1.br.C; }
	byte get_d () { return z.R1.br.D; }
	byte get_e () { return z.R1.br.E; }
	byte get_h () { return z.R1.br.H; }
	byte get_l () { return z.R1.br.L; }

	void set_a (byte v) { z.R1.br.A= v; }
	void set_f (byte v) { z.R1.br.F= v; }
	void set_b (byte v) { z.R1.br.B= v; }
	void set_c (byte v) { z.R1.br.C= v; }
	void set_d (byte v) { z.R1.br.D= v; }
	void set_e (byte v) { z.R1.br.E= v; }
	void set_h (byte v) { z.R1.br.H= v; }
	void set_l (byte v) { z.R1.br.L= v; }

	ushort get_af () { return z.R1.wr.AF; }
	ushort get_bc () { return z.R1.wr.BC; }
	ushort get_de () { return z.R1.wr.DE; }
	ushort get_hl () { return z.R1.wr.HL; }
	ushort get_ix () { return z.R1.wr.IX; }
	ushort get_iy () { return z.R1.wr.IY; }
	ushort get_sp () { return z.R1.wr.SP; }
	ushort get_pc () { return z.PC; }

	void set_af (ushort v) { z.R1.wr.AF= v; }
	void set_bc (ushort v) { z.R1.wr.BC= v; }
	void set_de (ushort v) { z.R1.wr.DE= v; }
	void set_hl (ushort v) { z.R1.wr.HL= v; }
	void set_ix (ushort v) { z.R1.wr.IX= v; }
	void set_iy (ushort v) { z.R1.wr.IY= v; }
	void set_sp (ushort v) { z.R1.wr.SP= v; }
	void set_pc (ushort v) { z.PC= v; }

	#else

private:
	// Virtual functions required for ImcZ80.
	int interrupt ()
	{
		return 0;
	}
	unsigned int out (unsigned long time, byte hi, byte lo, byte val)
	{
		// Returns the number of wait states.
		return 0;
	}
	unsigned int in (unsigned long time,
		unsigned long hi, unsigned long lo)
	{
		// Returns the value readed from the port plus the
		// number of wait states << 8.
		return 0;
	}
	void EDFE ()
	{ }
	void EDFD ()
	{ }
	void startwatch (int flag)
	{ }
	long stopwatch ()
	{
		return 0;
	}

public:
	// Registers access.

	//typedef unsigned short ushort;
	// Already defined.

	static byte flag_c (byte f) { return f & FLG_C; }
	static byte flag_z (byte f) { return f & FLG_Z; }
	static byte flag_s (byte f) { return f & FLG_S; }

	byte get_a () { return a; }
	byte get_f () { return f; }
	byte get_b () { return b; }
	byte get_c () { return c; }
	byte get_d () { return d; }
	byte get_e () { return e; }
	byte get_h () { return h; }
	byte get_l () { return l; }

	void set_a (byte v) { a= v; }
	void set_f (byte v) { f= v; }
	void set_b (byte v) { b= v; }
	void set_c (byte v) { c= v; }
	void set_d (byte v) { d= v; }
	void set_e (byte v) { e= v; }
	void set_h (byte v) { h= v; }
	void set_l (byte v) { l= v; }

	ushort get_af () { return (a << 8) | f; }
	ushort get_bc () { return (b << 8) | c; }
	ushort get_de () { return (d << 8) | e; }
	ushort get_hl () { return (h << 8) | l; }
	ushort get_ix () { return ix; }
	ushort get_iy () { return iy; }
	ushort get_sp () { return sp; }
	ushort get_pc () { return pc; }

	void set_af (ushort v) { a= (v >> 8); f= v & 0xFF; }
	void set_bc (ushort v) { b= (v >> 8); c= v & 0xFF; }
	void set_de (ushort v) { d= (v >> 8); e= v & 0xFF; }
	void set_hl (ushort v) { h= (v >> 8); l= v & 0xFF; }
	void set_ix (ushort v) { ix= v; }
	void set_iy (ushort v) { iy= v; }
	void set_sp (ushort v) { sp= v; }
	void set_pc (ushort v) { pc= v; }

private:
	// Memory access
	byte fetch (word x)
	{
		return mem [x];
	}
	void store (word x, byte y)
	{
		mem [x]= y;
	}

	#endif

public:
	Cpu ();
	virtual ~Cpu ();

	#if defined USE_ImcZ80 && defined DISASM_WITH_libz80
	void aux_set_pc (ushort npc)
	{ z.PC= npc; }
	#endif

	#if ! defined USE_ImcZ80 || defined DISASM_WITH_libz80
	void disassembly (std::ostream & out)
	{
		char hexbuf [128], buf [128];
		Z80Debug (& z, hexbuf, buf);
		out << 
			": " << hexbuf <<
			' ' << buf;
	}
	#endif

	void execute ()
	{
		#ifndef USE_ImcZ80

		Z80Execute (& z);

		#else

		ImcZ80::execute ();

		#endif
	}
	void reset ()
	{
		#ifdef USE_ImcZ80
		ImcZ80::reset ();
		#endif
	}
};

// End of cpu.h
